@@ -1,4 +1,4 @@
-/* $Id: milter-regex.c,v 1.44 2004/03/19 19:21:40 dhartmei Exp $ */
+/* $Id: milter-regex.c,v 1.49 2005/03/06 11:42:53 dhartmei Exp $ */
 
 /*
  * Copyright (c) 2003-2004 Daniel Hartmeier
@@ -30,7 +30,7 @@
  *
  */
 
-static const char rcsid[] = "$Id: milter-regex.c,v 1.44 2004/03/19 19:21:40 dhartmei Exp $";
+static const char rcsid[] = "$Id: milter-regex.c,v 1.49 2005/03/06 11:42:53 dhartmei Exp $";
 
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -63,10 +63,12 @@ struct context {
 	int		*res;
 	char		 buf[2048];	/* longer body lines are wrapped */
 	unsigned	 pos;		/* write position within buf */
-	char		 addr[64];
+	char		 host_name[128];
+	char		 host_addr[64];
 	char		 hdr_from[128];
 	char		 hdr_to[128];
 	char		 hdr_subject[128];
+	char		*quarantine;
 };
 
 static sfsistat		 setreply(SMFICTX *, struct context *,
@@ -80,7 +82,6 @@ static sfsistat		 cb_header(SMFICTX *, char *, char *);
 static sfsistat		 cb_eoh(SMFICTX *);
 static sfsistat		 cb_body(SMFICTX *, u_char *, size_t);
 static sfsistat		 cb_eom(SMFICTX *);
-static sfsistat		 cb_abort(SMFICTX *);
 static sfsistat		 cb_close(SMFICTX *);
 static void		 usage(const char *);
 static void		 msg(int, struct context *, const char *, ...);
@@ -137,37 +138,48 @@ daemon(int nochdir, int noclose)
 static sfsistat
 setreply(SMFICTX *ctx, struct context *context, const struct action *action)
 {
+	int result = SMFIS_CONTINUE;
+
 	switch (action->type) {
-	case SMFIS_REJECT:
+	case ACTION_REJECT:
 		msg(LOG_NOTICE, context, "REJECT: %s, From: %s, To: %s, "
 		    "Subject: %s", action->msg, context->hdr_from,
 		    context->hdr_to, context->hdr_subject);
+		result = SMFIS_REJECT;
 		break;
-	case SMFIS_TEMPFAIL:
+	case ACTION_TEMPFAIL:
 		msg(LOG_NOTICE, context, "TEMPFAIL: %s, From: %s, To: %s, "
 		    "Subject: %s", action->msg, context->hdr_from,
 		    context->hdr_to, context->hdr_subject);
+		result = SMFIS_TEMPFAIL;
 		break;
-	case SMFIS_DISCARD:
+	case ACTION_QUARANTINE:
+		if (context->quarantine != NULL)
+			free(context->quarantine);
+		context->quarantine = strdup(action->msg);
+		break;
+	case ACTION_DISCARD:
 		msg(LOG_NOTICE, context, "DISCARD, From: %s, To: %s, "
 		    "Subject: %s", context->hdr_from, context->hdr_to,
 		    context->hdr_subject);
+		result = SMFIS_DISCARD;
 		break;
-	case SMFIS_ACCEPT:
+	case ACTION_ACCEPT:
 		msg(LOG_INFO, context, "ACCEPT, From: %s, To: %s, "
 		    "Subject: %s", context->hdr_from, context->hdr_to,
 		    context->hdr_subject);
-		return (action->type);
+		result = SMFIS_ACCEPT;
+		break;
 	}
-	if (action->type == SMFIS_REJECT &&
+	if (action->type == ACTION_REJECT &&
 	    smfi_setreply(ctx, RCODE_REJECT, XCODE_REJECT,
 	    (char *)action->msg) != MI_SUCCESS)
 		msg(LOG_ERR, context, "smfi_setreply");
-	if (action->type == SMFIS_TEMPFAIL &&
+	if (action->type == ACTION_TEMPFAIL &&
 	    smfi_setreply(ctx, RCODE_TEMPFAIL, XCODE_TEMPFAIL,
 	    (char *)action->msg) != MI_SUCCESS)
 		msg(LOG_ERR, context, "smfi_setreply");
-	return (action->type);
+	return (result);
 }
 
 static struct ruleset *
@@ -231,8 +243,6 @@ static sfsistat
 cb_connect(SMFICTX *ctx, char *name, _SOCK_ADDR *sa)
 {
 	struct context *context;
-	const struct action *action;
-	char host[64];
 
 	context = calloc(1, sizeof(*context));
 	if (context == NULL) {
@@ -259,13 +269,15 @@ cb_connect(SMFICTX *ctx, char *name, _SOCK_ADDR *sa)
 	}
 	context->rs->refcnt++;
 
-	strlcpy(host, "unknown", sizeof(host));
+	strlcpy(context->host_name, name, sizeof(context->host_name));
+	strlcpy(context->host_addr, "unknown", sizeof(context->host_addr));
 	switch (sa->sa_family) {
 	case AF_INET: {
 		struct sockaddr_in *sin = (struct sockaddr_in *)sa;
 
-		if (inet_ntop(AF_INET, &sin->sin_addr.s_addr, host,
-		    sizeof(host)) == NULL)
+		if (inet_ntop(AF_INET, &sin->sin_addr.s_addr,
+		    context->host_addr, sizeof(context->host_addr)) ==
+		    NULL)
 			msg(LOG_ERR, NULL, "cb_connect: inet_ntop: %s",
 			    strerror(errno));
 		break;
@@ -273,21 +285,16 @@ cb_connect(SMFICTX *ctx, char *name, _SOCK_ADDR *sa)
 	case AF_INET6: {
 		struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)sa;
 
-		if (inet_ntop(AF_INET6, &sin6->sin6_addr, host,
-		    sizeof(host)) == NULL)
+		if (inet_ntop(AF_INET6, &sin6->sin6_addr,
+		    context->host_addr, sizeof(context->host_addr)) ==
+		    NULL)
 			msg(LOG_ERR, NULL, "cb_connect: inet_ntop: %s",
 			    strerror(errno));
 		break;
 	}
 	}
-	strlcpy(context->addr, host, sizeof(context->addr));
-	msg(LOG_DEBUG, context, "cb_connect('%s', '%s')", name, host);
-	if ((action = eval_cond(context->rs, context->res, COND_CONNECT,
-	    name, host)) != NULL)
-		return (setreply(ctx, context, action));
-	if ((action = eval_end(context->rs, context->res, COND_CONNECT)) !=
-	    NULL)
-		return (setreply(ctx, context, action));
+	msg(LOG_DEBUG, context, "cb_connect('%s', '%s')",
+	    context->host_name, context->host_addr);
 	return (SMFIS_CONTINUE);
 }
 
@@ -301,8 +308,17 @@ cb_helo(SMFICTX *ctx, char *arg)
 		msg(LOG_ERR, NULL, "cb_helo: smfi_getpriv");
 		return (SMFIS_ACCEPT);
 	}
-	msg(LOG_DEBUG, context, "cb_helo('%s')", arg);
 	/* multiple HELO imply RSET in sendmail */
+	/* evaluate connect arguments here, because we can't call */
+	/* setreply from cb_connect */
+	eval_clear(context->rs, context->res, COND_CONNECT);
+	if ((action = eval_cond(context->rs, context->res, COND_CONNECT,
+	    context->host_name, context->host_addr)) != NULL)
+		return (setreply(ctx, context, action));
+	if ((action = eval_end(context->rs, context->res, COND_CONNECT)) !=
+	    NULL)
+		return (setreply(ctx, context, action));
+	msg(LOG_DEBUG, context, "cb_helo('%s')", arg);
 	eval_clear(context->rs, context->res, COND_HELO);
 	if ((action = eval_cond(context->rs, context->res, COND_HELO,
 	    arg, NULL)) != NULL)
@@ -347,12 +363,17 @@ cb_envrcpt(SMFICTX *ctx, char **args)
 		msg(LOG_ERR, NULL, "cb_envrcpt: smfi_getpriv");
 		return (SMFIS_ACCEPT);
 	}
+	/* multiple RCPT TO: possible */
+	eval_clear(context->rs, context->res, COND_ENVRCPT);
 	if (*args != NULL) {
 		msg(LOG_DEBUG, context, "cb_envrcpt('%s')", *args);
 		if ((action = eval_cond(context->rs, context->res, COND_ENVRCPT,
 		    *args, NULL)) != NULL)
 			return (setreply(ctx, context, action));
 	}
+	if ((action = eval_end(context->rs, context->res, COND_ENVRCPT)) !=
+	    NULL)
+		return (setreply(ctx, context, action));
 	return (SMFIS_CONTINUE);
 }
 
@@ -366,9 +387,6 @@ cb_header(SMFICTX *ctx, char *name, char *value)
 		msg(LOG_ERR, context, "cb_header: smfi_getpriv");
 		return (SMFIS_ACCEPT);
 	}
-	if ((action = eval_end(context->rs, context->res, COND_ENVRCPT)) !=
-	    NULL)
-		return (setreply(ctx, context, action));
 	msg(LOG_DEBUG, context, "cb_header('%s', '%s')", name, value);
 	if (!strcasecmp(name, "From"))
 		strlcpy(context->hdr_from, value, sizeof(context->hdr_from));
@@ -438,6 +456,7 @@ cb_eom(SMFICTX *ctx)
 {
 	struct context *context;
 	const struct action *action;
+	int result = SMFIS_ACCEPT;
 
 	if ((context = (struct context *)smfi_getpriv(ctx)) == NULL) {
 		msg(LOG_ERR, NULL, "cb_eom: smfi_getpriv");
@@ -446,27 +465,19 @@ cb_eom(SMFICTX *ctx)
 	msg(LOG_DEBUG, context, "cb_eom()");
 	if ((action = eval_end(context->rs, context->res, COND_BODY)) !=
 	    NULL)
-		return (setreply(ctx, context, action));
+		result = setreply(ctx, context, action);
 	else
 		msg(LOG_DEBUG, context, "ACCEPT, From: %s, To: %s, "
 		    "Subject: %s", context->hdr_from, context->hdr_to,
 		    context->hdr_subject);
-	return (SMFIS_CONTINUE);
-}
-
-static sfsistat
-cb_abort(SMFICTX *ctx)
-{
-	struct context *context;
-
-	if ((context = (struct context *)smfi_getpriv(ctx)) == NULL) {
-		msg(LOG_ERR, NULL, "cb_abort: smfi_getpriv");
-		return (SMFIS_ACCEPT);
+	if (context->quarantine != NULL) {
+		msg(LOG_NOTICE, context, "QUARANTINE: %s, From: %s, To: %s, "
+		    "Subject: %s", action->msg, context->hdr_from,
+		    context->hdr_to, context->hdr_subject);
+		if (smfi_quarantine(ctx, context->quarantine) != MI_SUCCESS)
+			msg(LOG_ERR, context, "cb_eom: smfi_quarantine");
 	}
-	msg(LOG_DEBUG, context, "cb_abort()");
-	/* a RSET doesn't clear HELO in sendmail, but MAIL FROM */
-	eval_clear(context->rs, context->res, COND_ENVFROM);
-	return (SMFIS_CONTINUE);
+	return (result);
 }
 
 static sfsistat
@@ -479,6 +490,8 @@ cb_close(SMFICTX *ctx)
 	if (context != NULL) {
 		smfi_setpriv(ctx, NULL);
 		free(context->res);
+		if (context->quarantine != NULL)
+			free(context->quarantine);
 		context->rs->refcnt--;
 		free(context);
 	}
@@ -488,7 +501,7 @@ cb_close(SMFICTX *ctx)
 struct smfiDesc smfilter = {
 	"milter-regex",	/* filter name */
 	SMFI_VERSION,	/* version code -- do not change */
-	SMFIF_ADDHDRS|SMFIF_CHGHDRS|SMFIF_ADDRCPT|SMFIF_DELRCPT, /* flags */
+	SMFIF_QUARANTINE, /* flags */
 	cb_connect,	/* connection info filter */
 	cb_helo,	/* SMTP HELO command filter */
 	cb_envfrom,	/* envelope sender filter */
@@ -497,7 +510,7 @@ struct smfiDesc smfilter = {
 	cb_eoh,		/* end of header */
 	cb_body,	/* body block */
 	cb_eom,		/* end of message */
-	cb_abort,	/* message aborted */
+	NULL,		/* message aborted */
 	cb_close	/* connection cleanup */
 };
 
@@ -509,7 +522,7 @@ msg(int priority, struct context *context, const char *fmt, ...)
 
 	va_start(ap, fmt);
 	if (context != NULL)
-		snprintf(msg, sizeof(msg), "%s: ", context->addr);
+		snprintf(msg, sizeof(msg), "%s: ", context->host_addr);
 	else
 		msg[0] = 0;
 	vsnprintf(msg + strlen(msg), sizeof(msg) - strlen(msg), fmt, ap);
@@ -615,7 +628,7 @@ main(int argc, char **argv)
 		goto done;
 	}
 
-	if (eval_init(SMFIS_ACCEPT)) {
+	if (eval_init(ACTION_ACCEPT)) {
 		fprintf(stderr, "eval_init: failed\n");
 		goto done;
 	}
